/*****************************************************************************\
* (c) Copyright 2018-2020 CERN for the benefit of the LHCb Collaboration      *
*                                                                             *
* This software is distributed under the terms of the Apache License          *
* version 2 (Apache-2.0), copied verbatim in the file "LICENSE".              *
*                                                                             *
* In applying this licence, CERN does not waive the privileges and immunities *
* granted to it by virtue of its status as an Intergovernmental Organization  *
* or submit itself to any jurisdiction.                                       *
\*****************************************************************************/
#pragma once

#include "Store.cuh"
#include "Configuration.h"
#include "Logger.h"
#include <utility>
#include <type_traits>
#include <AlgorithmDB.h>
#include "nlohmann/json.hpp"

// use constexpr flag to enable/disable contracts
#ifdef ENABLE_CONTRACTS
constexpr bool contracts_enabled = true;
#else
constexpr bool contracts_enabled = false;
#endif

namespace Allen {
  struct ScheduledSequence {
    std::vector<Allen::TypeErasedAlgorithm> sequence;
    std::tuple<std::vector<LifetimeDependencies>, std::vector<LifetimeDependencies>> dependencies;

    ScheduledSequence(const ConfiguredSequence& configuration)
    {
      const auto& [configured_algorithms, configured_arguments, sequence_arguments, arg_deps] = configuration;
      sequence = instantiate_sequence(configured_algorithms);
      dependencies = calculate_lifetime_dependencies(sequence_arguments, arg_deps, configured_arguments, sequence);
    }

    // Configure constants for algorithms in the sequence
    void configure_algorithms(const std::map<std::string, std::map<std::string, nlohmann::json>>& config)
    {
      for (auto& algorithm : sequence) {
        auto c = config.find(algorithm.name());
        if (c != config.end()) algorithm.set_properties(c->second);
      }
    }

    // Configure constants for algorithms in the sequence
    void initialize_algorithms()
    {
      for (auto& algorithm : sequence) {
        // Invoke void initialize() const, if it exists
        algorithm.init();
      }
    }

    void update_algorithms(const Constants& constants)
    {
      for (auto& algorithm : sequence) {
        algorithm.update(constants);
      }
    }

    static std::vector<Allen::TypeErasedAlgorithm> instantiate_sequence(
      const std::vector<ConfiguredAlgorithm>& configured_algorithms)
    {
      std::vector<Allen::TypeErasedAlgorithm> sequence;
      sequence.reserve(configured_algorithms.size());
      for (const auto& alg : configured_algorithms) {
        sequence.emplace_back(instantiate_allen_algorithm(alg));
      }
      return sequence;
    }

  private:
    std::tuple<std::vector<LifetimeDependencies>, std::vector<LifetimeDependencies>> calculate_lifetime_dependencies(
      const std::vector<ConfiguredAlgorithmArguments>& sequence_arguments,
      const ArgumentDependencies& argument_dependencies,
      const std::vector<ConfiguredArgument>& configured_arguments,
      const std::vector<Allen::TypeErasedAlgorithm>& sequence)
    {
      std::vector<LifetimeDependencies> in_deps;
      std::vector<LifetimeDependencies> out_deps;
      std::vector<std::string> temp_arguments;

      // Make map of configured arguments
      std::map<std::string, std::string> configured_arguments_map;
      for (const auto& conf_arg : configured_arguments) {
        configured_arguments_map[conf_arg.name] = conf_arg.scope;
      }

      const auto argument_in = [](const std::string& arg, const auto& args) {
        return std::find(std::begin(args), std::end(args), arg) != std::end(args);
      };

      const auto argument_in_map = [](const std::string& arg, const auto& args) {
        return args.find(arg) != std::end(args);
      };

      auto seq_args = sequence_arguments;

      // Add all dependencies from all SelectionAlgorithms to in_deps of algorithm gather_selections
      std::set<std::string> selection_arguments;
      for (unsigned i = 0; i < seq_args.size(); ++i) {
        if (sequence[i].scope() == "SelectionAlgorithm") {
          for (const auto& arg : seq_args[i].arguments) {
            selection_arguments.insert(arg);
          }
        }
        if (sequence[i].scope() == "BarrierAlgorithm") {
          for (const auto& arg : selection_arguments) {
            seq_args[i].arguments.push_back(arg);
          }
        }
      }

      for (unsigned i = 0; i < seq_args.size(); ++i) {
        // Calculate out_dep for this algorithm
        LifetimeDependencies out_dep;
        std::vector<std::string> next_temp_arguments;

        for (const auto& arg : temp_arguments) {
          bool arg_can_be_freed = true;

          // The argument can be freed only if it is not host
          if (configured_arguments_map[arg] == "host") {
            arg_can_be_freed = false;
          }

          for (unsigned j = i; j < seq_args.size(); ++j) {
            const auto& alg = seq_args[j];
            if (argument_in(arg, alg.arguments)) {
              arg_can_be_freed = false;
            }

            // dependencies
            for (const auto& alg_arg : alg.arguments) {
              if (
                argument_in_map(alg_arg, argument_dependencies) &&
                argument_in(arg, argument_dependencies.at(alg_arg))) {
                arg_can_be_freed = false;
                break;
              }
            }

            // input aggregates
            for (const auto& input_aggregate : alg.input_aggregates) {
              if (argument_in(arg, input_aggregate)) {
                arg_can_be_freed = false;
                break;
              }
            }

            if (!arg_can_be_freed) {
              break;
            }
          }

          if (arg_can_be_freed) {
            out_dep.arguments.push_back(arg);
          }
          else {
            next_temp_arguments.push_back(arg);
          }
        }
        out_deps.emplace_back(out_dep);

        // Update temp_arguments
        temp_arguments = next_temp_arguments;

        // Calculate in_dep for this algorithm
        LifetimeDependencies in_dep;
        for (const auto& arg : seq_args[i].arguments) {
          if (!argument_in(arg, temp_arguments)) {
            temp_arguments.push_back(arg);
            in_dep.arguments.push_back(arg);
          }
        }
        in_deps.emplace_back(in_dep);
      }

      return {in_deps, out_deps};
    }
  };
} // namespace Allen

class Scheduler {
  const Allen::ScheduledSequence* m_sched_seq;
  Allen::Store::UnorderedStore m_store;
  std::vector<std::any> m_sequence_ref_stores;
  std::vector<LifetimeDependencies> m_in_dependencies;
  std::vector<LifetimeDependencies> m_out_dependencies;
  bool do_print = false;

public:
  Scheduler(
    const ConfiguredSequence& configuration,
    const Allen::ScheduledSequence& sched_seq,
    const bool param_do_print,
    const size_t device_requested_mb,
    const unsigned required_memory_alignment)
  {
    auto& [configured_algorithms, configured_arguments, sequence_arguments, arg_deps] = configuration;
    assert(configured_algorithms.size() == sequence_arguments.size());

    m_sched_seq = &sched_seq;

    // Create and populate store
    initialize_store(configured_arguments, sequence_arguments);

    // Calculate in and out dependencies of defined sequence
    std::tie(m_in_dependencies, m_out_dependencies) = sched_seq.dependencies;

    // Create ArgumentRefManager of each algorithm
    for (unsigned i = 0; i < m_sched_seq->sequence.size(); ++i) {
      // Generate store references for each algorithm's configured arguments
      auto [alg_arguments, alg_input_aggregates] = generate_algorithm_store_ref(sequence_arguments[i]);
      m_sequence_ref_stores.emplace_back(
        m_sched_seq->sequence[i].create_ref_store(alg_arguments, alg_input_aggregates, m_store));
    }

    assert(configured_algorithms.size() == m_sched_seq->sequence.size());
    assert(configured_algorithms.size() == m_sequence_ref_stores.size());
    assert(configured_algorithms.size() == m_in_dependencies.size());
    assert(configured_algorithms.size() == m_out_dependencies.size());

    do_print = param_do_print;

    // Reserve memory
    m_store.reserve_memory_device(device_requested_mb, required_memory_alignment);
  }

  Scheduler(const Scheduler&) = delete;
  Scheduler& operator=(const Scheduler&) = delete;
  Scheduler(Scheduler&&) = delete;
  Scheduler& operator=(Scheduler&&) = delete;

  /**
   * @brief Initializes the store with the configured arguments
   */
  void initialize_store(
    const std::vector<ConfiguredArgument>&,
    const std::vector<ConfiguredAlgorithmArguments>& configured_algorithm_arguments)
  {
    assert(m_sched_seq->sequence.size() == configured_algorithm_arguments.size());
    for (unsigned i = 0; i < m_sched_seq->sequence.size(); ++i) {
      m_sched_seq->sequence[i].emplace_output_arguments(configured_algorithm_arguments[i].arguments, m_store);
    }
  }

  /**
   * @brief Generate the store ref of an algorithm
   */
  std::tuple<
    std::vector<std::reference_wrapper<Allen::Store::BaseArgument>>,
    std::vector<std::vector<std::reference_wrapper<Allen::Store::BaseArgument>>>>
  generate_algorithm_store_ref(const ConfiguredAlgorithmArguments& configured_alg_arguments)
  {
    std::vector<std::reference_wrapper<Allen::Store::BaseArgument>> store_ref;
    std::vector<std::vector<std::reference_wrapper<Allen::Store::BaseArgument>>> input_aggregates;

    for (const auto& argument : configured_alg_arguments.arguments) {
      store_ref.push_back(m_store.at(argument));
    }

    for (const auto& conf_input_aggregate : configured_alg_arguments.input_aggregates) {
      std::vector<std::reference_wrapper<Allen::Store::BaseArgument>> input_aggregate;
      for (const auto& argument : conf_input_aggregate) {
        input_aggregate.push_back(m_store.at(argument));
      }
      input_aggregates.emplace_back(input_aggregate);
    }

    return {store_ref, input_aggregates};
  }

  /**
   * @brief Free the memory managers.
   */
  void free_all() { m_store.free_all(); }

  // Return constants for algorithms in the sequence
  auto get_algorithm_configuration() const
  {
    std::map<std::string, std::map<std::string, nlohmann::json>> config;
    for (unsigned i = 0; i < m_sched_seq->sequence.size(); ++i) {
      get_configuration(m_sched_seq->sequence[i], config);
    }
    return config;
  }

  void print_sequence() const
  {
    info_cout << "\nSequence:\n";
    for (const auto& alg : m_sched_seq->sequence) {
      info_cout << "  " << alg.name() << "\n";
    }
    info_cout << "\n";
  }

  bool contains_validation_algorithms() const
  {
    for (const auto& alg : m_sched_seq->sequence) {
      if (alg.scope() == "ValidationAlgorithm") {
        return true;
      }
    }
    return false;
  }

  //  Runs a sequence of algorithms.
  void run(
    const RuntimeOptions& runtime_options,
    const Constants& constants,
    Allen::Store::PersistentStore* persistent_store,
    const Allen::Context& context)
  {
    m_store.set_persistent_store(persistent_store);
    for (unsigned i = 0; i < m_sched_seq->sequence.size(); ++i) {
      run(
        m_sched_seq->sequence[i],
        m_sequence_ref_stores[i],
        m_in_dependencies[i],
        m_out_dependencies[i],
        m_store,
        runtime_options,
        constants,
        context,
        do_print);
    }
    m_store.set_persistent_store_map();
  }

private:
  static void get_configuration(
    const Allen::TypeErasedAlgorithm& algorithm,
    std::map<std::string, std::map<std::string, nlohmann::json>>& config)
  {
    config.emplace(algorithm.name(), algorithm.get_properties());
  }

  static void setup(
    const Allen::TypeErasedAlgorithm& algorithm,
    const LifetimeDependencies& in_dependencies,
    const LifetimeDependencies& out_dependencies,
    Allen::Store::UnorderedStore& store,
    bool do_print)
  {
    /**
     * @brief Runs a step of the scheduler and determines
     *        the offset for each argument.
     *
     *        The sequence is asserted at compile time to run the
     *        expected iteration and reserve the expected types.
     *
     *        This function should always be invoked, even when it is
     *        known there are no tags to reserve or free on this step.
     */
    if (do_print) {
      info_cout << "Sequence step \"" << algorithm.name() << "\":\n";
    }

    // Free arguments in OutDependencies
    for (const auto& arg : out_dependencies.arguments) {
      store.free(arg);
    }

    // Reserve all arguments in InDependencies
    for (const auto& arg : in_dependencies.arguments) {
      store.put(arg);
    }

    // Print memory manager state
    if (do_print) {
      store.print_memory_manager_states();
    }
  }

  static void run(
    const Allen::TypeErasedAlgorithm& algorithm,
    std::any& argument_ref_manager,
    const LifetimeDependencies& in_dependencies,
    const LifetimeDependencies& out_dependencies,
    Allen::Store::UnorderedStore& store,
    const RuntimeOptions& runtime_options,
    const Constants& constants,
    const Allen::Context& context,
    bool do_print)
  {
    // Sets the arguments sizes
    algorithm.set_arguments_size(argument_ref_manager, runtime_options, constants);

    // Setup algorithm, reserving / freeing memory buffers
    setup(algorithm, in_dependencies, out_dependencies, store, do_print);

    // Run preconditions
    if constexpr (contracts_enabled) {
      algorithm.run_preconditions(argument_ref_manager, runtime_options, constants, context);
    }

    try {
      // Invoke the algorithm
      algorithm.invoke(argument_ref_manager, runtime_options, constants, context);
    } catch (std::invalid_argument& e) {
      fprintf(stderr, "Execution of algorithm %s raised an exception\n", algorithm.name().c_str());
      throw e;
    }

    // Run postconditions
    if constexpr (contracts_enabled) {
      algorithm.run_postconditions(argument_ref_manager, runtime_options, constants, context);
    }
  }
};
